# mypy: allow-untyped-defs
import copy
import warnings
from typing import Any, Callable, Optional, Union

import torch
import torch.nn.functional as F
from torch import Tensor
from torch.nn.init import xavier_uniform_

from torch.nn.modules.activation import MultiheadAttention
from torch.nn.modules.container import ModuleList
from torch.nn.modules.dropout import Dropout
from torch.nn.modules.linear import Linear
from torch.nn.modules import Module
from torch.nn.modules.normalization import LayerNorm


__all__ = [
    "TransformerEncoder",
    "TransformerEncoderLayer",
]


def _generate_square_subsequent_mask(
    sz: int,
    device: Optional[torch.device] = None,
    dtype: Optional[torch.dtype] = None,
) -> Tensor:
    r"""Generate a square causal mask for the sequence.

    The masked positions are filled with float('-inf'). Unmasked positions are filled with float(0.0).
    """
    return torch.triu(
        torch.full((sz, sz), float("-inf"), dtype=dtype, device=device),
        diagonal=1,
    )


def _get_seq_len(src: Tensor, batch_first: bool) -> Optional[int]:
    if src.is_nested:
        return None
    else:
        src_size = src.size()
        if len(src_size) == 2:
            # unbatched: S, E
            return src_size[0]
        else:
            # batched: B, S, E if batch_first else S, B, E
            seq_len_pos = 1 if batch_first else 0
            return src_size[seq_len_pos]


class TransformerEncoder(Module):
    r"""TransformerEncoder is a stack of N encoder layers.

    This TransformerEncoder layer implements the original architecture described
    in the `Attention Is All You Need <https://arxiv.org/abs/1706.03762>`_ paper. The
    intent of this layer is as a reference implementation for foundational understanding
    and thus it contains only limited features relative to newer Transformer architectures.
    Given the fast pace of innovation in transformer-like architectures, we recommend
    exploring this `tutorial <https://pytorch.org/tutorials/intermediate/transformer_building_blocks.html>`_
    to build efficient layers from building blocks in core or using higher
    level libraries from the `PyTorch Ecosystem <https://landscape.pytorch.org/>`_.

    .. warning::
        All layers in the TransformerEncoder are initialized with the same parameters.
        It is recommended to manually initialize the layers after creating the TransformerEncoder instance.

    Args:
        encoder_layer: an instance of the TransformerEncoderLayer() class (required).
        num_layers: the number of sub-encoder-layers in the encoder (required).
        norm: the layer normalization component (optional).
        enable_nested_tensor: if True, input will automatically convert to nested tensor
            (and convert back on output). This will improve the overall performance of
            TransformerEncoder when padding rate is high. Default: ``True`` (enabled).

    Examples:
        >>> encoder_layer = nn.TransformerEncoderLayer(d_model=512, nhead=8)
        >>> transformer_encoder = nn.TransformerEncoder(encoder_layer, num_layers=6)
        >>> src = torch.rand(10, 32, 512)
        >>> out = transformer_encoder(src)
    """

    __constants__ = ["norm"]

    def __init__(
        self,
        encoder_layer: "TransformerEncoderLayer",
        num_layers: int,
        norm: Optional[Module] = None,
        enable_nested_tensor: bool = True,
        mask_check: bool = True,
    ) -> None:
        super().__init__()
        torch._C._log_api_usage_once(f"torch.nn.modules.{self.__class__.__name__}")
        self.layers = _get_clones(encoder_layer, num_layers)
        self.num_layers = num_layers
        self.norm = norm
        # this attribute saves the value providedat object construction
        self.enable_nested_tensor = enable_nested_tensor
        # this attribute controls whether nested tensors are used
        self.use_nested_tensor = enable_nested_tensor
        self.mask_check = mask_check

        enc_layer = "encoder_layer"
        why_not_sparsity_fast_path = ""
        if not isinstance(encoder_layer, torch.nn.TransformerEncoderLayer):
            why_not_sparsity_fast_path = f"{enc_layer} was not TransformerEncoderLayer"
        elif encoder_layer.norm_first:
            why_not_sparsity_fast_path = f"{enc_layer}.norm_first was True"
        elif not encoder_layer.self_attn.batch_first:
            why_not_sparsity_fast_path = (
                f"{enc_layer}.self_attn.batch_first was not True"
                + "(use batch_first for better inference performance)"
            )
        elif not encoder_layer.self_attn._qkv_same_embed_dim:
            why_not_sparsity_fast_path = (
                f"{enc_layer}.self_attn._qkv_same_embed_dim was not True"
            )
        elif encoder_layer.self_attn.in_proj_bias is None:
            why_not_sparsity_fast_path = f"{enc_layer}.self_attn was passed bias=False"
        elif not encoder_layer.activation_relu_or_gelu:
            why_not_sparsity_fast_path = (
                f"{enc_layer}.activation_relu_or_gelu was not True"
            )
        elif not (encoder_layer.norm1.eps == encoder_layer.norm2.eps):
            why_not_sparsity_fast_path = (
                f"{enc_layer}.norm1.eps was not equal to {enc_layer}.norm2.eps"
            )
        elif encoder_layer.self_attn.num_heads % 2 == 1:
            why_not_sparsity_fast_path = f"{enc_layer}.self_attn.num_heads is odd"

        if enable_nested_tensor and why_not_sparsity_fast_path:
            warnings.warn(
                f"enable_nested_tensor is True, but self.use_nested_tensor is False because {why_not_sparsity_fast_path}"
            )
            self.use_nested_tensor = False

    def forward(
        self,
        src: Tensor,
        mask: Optional[Tensor] = None,
        src_key_padding_mask: Optional[Tensor] = None,
        is_causal: Optional[bool] = None,
        output_attentions: bool = False,
    ) -> Tensor:
        r"""Pass the input through the encoder layers in turn.

        Args:
            src: the sequence to the encoder (required).
            mask: the mask for the src sequence (optional).
            src_key_padding_mask: the mask for the src keys per batch (optional).
            is_causal: If specified, applies a causal mask as ``mask``.
                Default: ``None``; try to detect a causal mask.
                Warning:
                ``is_causal`` provides a hint that ``mask`` is the
                causal mask. Providing incorrect hints can result in
                incorrect execution, including forward and backward
                compatibility.

        Shape:
            see the docs in :class:`~torch.nn.Transformer`.
        """
        src_key_padding_mask = F._canonical_mask(
            mask=src_key_padding_mask,
            mask_name="src_key_padding_mask",
            other_type=F._none_or_dtype(mask),
            other_name="mask",
            target_type=src.dtype,
        )

        mask = F._canonical_mask(
            mask=mask,
            mask_name="mask",
            other_type=None,
            other_name="",
            target_type=src.dtype,
            check_other=False,
        )

        output = src
        convert_to_nested = False
        first_layer = self.layers[0]
        src_key_padding_mask_for_layers = src_key_padding_mask
        why_not_sparsity_fast_path = ""
        str_first_layer = "self.layers[0]"
        batch_first = first_layer.self_attn.batch_first
        is_fastpath_enabled = torch.backends.mha.get_fastpath_enabled()

        if not is_fastpath_enabled:
            why_not_sparsity_fast_path = (
                "torch.backends.mha.get_fastpath_enabled() was not True"
            )
        elif not hasattr(self, "use_nested_tensor"):
            why_not_sparsity_fast_path = "use_nested_tensor attribute not present"
        elif not self.use_nested_tensor:
            why_not_sparsity_fast_path = (
                "self.use_nested_tensor (set in init) was not True"
            )
        elif first_layer.training:
            why_not_sparsity_fast_path = f"{str_first_layer} was in training mode"
        elif not src.dim() == 3:
            why_not_sparsity_fast_path = (
                f"input not batched; expected src.dim() of 3 but got {src.dim()}"
            )
        elif src_key_padding_mask is None:
            why_not_sparsity_fast_path = "src_key_padding_mask was None"
        elif (
            (not hasattr(self, "mask_check")) or self.mask_check
        ) and not torch._nested_tensor_from_mask_left_aligned(
            src, src_key_padding_mask.logical_not()
        ):
            why_not_sparsity_fast_path = "mask_check enabled, and src and src_key_padding_mask was not left aligned"
        elif output.is_nested:
            why_not_sparsity_fast_path = "NestedTensor input is not supported"
        elif mask is not None:
            why_not_sparsity_fast_path = (
                "src_key_padding_mask and mask were both supplied"
            )
        elif torch.is_autocast_enabled():
            why_not_sparsity_fast_path = "autocast is enabled"

        if not why_not_sparsity_fast_path:
            tensor_args = (
                src,
                first_layer.self_attn.in_proj_weight,
                first_layer.self_attn.in_proj_bias,
                first_layer.self_attn.out_proj.weight,
                first_layer.self_attn.out_proj.bias,
                first_layer.norm1.weight,
                first_layer.norm1.bias,
                first_layer.norm2.weight,
                first_layer.norm2.bias,
                first_layer.linear1.weight,
                first_layer.linear1.bias,
                first_layer.linear2.weight,
                first_layer.linear2.bias,
            )
            _supported_device_type = [
                "cpu",
                "cuda",
                torch.utils.backend_registration._privateuse1_backend_name,
            ]
            if torch.overrides.has_torch_function(tensor_args):
                why_not_sparsity_fast_path = "some Tensor argument has_torch_function"
            elif src.device.type not in _supported_device_type:
                why_not_sparsity_fast_path = (
                    f"src device is neither one of {_supported_device_type}"
                )
            elif torch.is_grad_enabled() and any(x.requires_grad for x in tensor_args):
                why_not_sparsity_fast_path = (
                    "grad is enabled and at least one of query or the "
                    "input/output projection weights or biases requires_grad"
                )

            if (not why_not_sparsity_fast_path) and (src_key_padding_mask is not None):
                convert_to_nested = True
                output = torch._nested_tensor_from_mask(
                    output, src_key_padding_mask.logical_not(), mask_check=False
                )
                src_key_padding_mask_for_layers = None

        seq_len = _get_seq_len(src, batch_first)
        is_causal = _detect_is_causal_mask(mask, is_causal, seq_len)

        attn_maps = []
        for mod in self.layers:
            if output_attentions:
                output, attn_map = mod(
                    output,
                    src_mask=mask,
                    src_key_padding_mask=src_key_padding_mask_for_layers,
                    is_causal=is_causal,
                    output_attentions=True,
                )
                attn_maps.append(attn_map)
            else:
                output = mod(
                    output,
                    src_mask=mask,
                    is_causal=is_causal,
                    src_key_padding_mask=src_key_padding_mask_for_layers,
                )

        if convert_to_nested:
            output = output.to_padded_tensor(0.0, src.size())

        if self.norm is not None:
            output = self.norm(output)

        if output_attentions:
            return output, attn_maps
        else:
            return output


class TransformerEncoderLayer(Module):
    r"""TransformerEncoderLayer is made up of self-attn and feedforward network.

    This TransformerEncoderLayer implements the original architecture described
    in the `Attention Is All You Need <https://arxiv.org/abs/1706.03762>`_ paper. The
    intent of this layer is as a reference implementation for foundational understanding
    and thus it contains only limited features relative to newer Transformer architectures.
    Given the fast pace of innovation in transformer-like architectures, we recommend
    exploring this `tutorial <https://pytorch.org/tutorials/intermediate/transformer_building_blocks.html>`_
    to build efficient layers from building blocks in core or using higher
    level libraries from the `PyTorch Ecosystem <https://landscape.pytorch.org/>`_.

    TransformerEncoderLayer can handle either traditional torch.tensor inputs,
    or Nested Tensor inputs.  Derived classes are expected to similarly accept
    both input formats.  (Not all combinations of inputs are currently
    supported by TransformerEncoderLayer while Nested Tensor is in prototype
    state.)

    If you are implementing a custom layer, you may derive it either from
    the Module or TransformerEncoderLayer class.  If your custom layer
    supports both torch.Tensors and Nested Tensors inputs, make its
    implementation a derived class of TransformerEncoderLayer. If your custom
    Layer supports only torch.Tensor inputs, derive its implementation from
    Module.

    Args:
        d_model: the number of expected features in the input (required).
        nhead: the number of heads in the multiheadattention models (required).
        dim_feedforward: the dimension of the feedforward network model (default=2048).
        dropout: the dropout value (default=0.1).
        activation: the activation function of the intermediate layer, can be a string
            ("relu" or "gelu") or a unary callable. Default: relu
        layer_norm_eps: the eps value in layer normalization components (default=1e-5).
        batch_first: If ``True``, then the input and output tensors are provided
            as (batch, seq, feature). Default: ``False`` (seq, batch, feature).
        norm_first: if ``True``, layer norm is done prior to attention and feedforward
            operations, respectively. Otherwise it's done after. Default: ``False`` (after).
        bias: If set to ``False``, ``Linear`` and ``LayerNorm`` layers will not learn an additive
            bias. Default: ``True``.

    Examples:
        >>> encoder_layer = nn.TransformerEncoderLayer(d_model=512, nhead=8)
        >>> src = torch.rand(10, 32, 512)
        >>> out = encoder_layer(src)

    Alternatively, when ``batch_first`` is ``True``:
        >>> encoder_layer = nn.TransformerEncoderLayer(
        ...     d_model=512, nhead=8, batch_first=True
        ... )
        >>> src = torch.rand(32, 10, 512)
        >>> out = encoder_layer(src)

    Fast path:
        forward() will use a special optimized implementation described in
        `FlashAttention: Fast and Memory-Efficient Exact Attention with IO-Awareness`_ if all of the following
        conditions are met:

        - Either autograd is disabled (using ``torch.inference_mode`` or ``torch.no_grad``) or no tensor
          argument ``requires_grad``
        - training is disabled (using ``.eval()``)
        - batch_first is ``True`` and the input is batched (i.e., ``src.dim() == 3``)
        - activation is one of: ``"relu"``, ``"gelu"``, ``torch.functional.relu``, or ``torch.functional.gelu``
        - at most one of ``src_mask`` and ``src_key_padding_mask`` is passed
        - if src is a `NestedTensor <https://pytorch.org/docs/stable/nested.html>`_, neither ``src_mask``
          nor ``src_key_padding_mask`` is passed
        - the two ``LayerNorm`` instances have a consistent ``eps`` value (this will naturally be the case
          unless the caller has manually modified one without modifying the other)

        If the optimized implementation is in use, a
        `NestedTensor <https://pytorch.org/docs/stable/nested.html>`_ can be
        passed for ``src`` to represent padding more efficiently than using a padding
        mask. In this case, a `NestedTensor <https://pytorch.org/docs/stable/nested.html>`_ will be
        returned, and an additional speedup proportional to the fraction of the input that
        is padding can be expected.

        .. _`FlashAttention: Fast and Memory-Efficient Exact Attention with IO-Awareness`:
         https://arxiv.org/abs/2205.14135

    """

    __constants__ = ["norm_first"]

    def __init__(
        self,
        d_model: int,
        nhead: int,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: Union[str, Callable[[Tensor], Tensor]] = F.relu,
        layer_norm_eps: float = 1e-5,
        batch_first: bool = False,
        norm_first: bool = False,
        bias: bool = True,
        device=None,
        dtype=None,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.self_attn = MultiheadAttention(
            d_model,
            nhead,
            dropout=dropout,
            bias=bias,
            batch_first=batch_first,
            **factory_kwargs,
        )
        # Implementation of Feedforward model
        self.linear1 = Linear(d_model, dim_feedforward, bias=bias, **factory_kwargs)
        self.dropout = Dropout(dropout)
        self.linear2 = Linear(dim_feedforward, d_model, bias=bias, **factory_kwargs)

        self.norm_first = norm_first
        self.norm1 = LayerNorm(d_model, eps=layer_norm_eps, bias=bias, **factory_kwargs)
        self.norm2 = LayerNorm(d_model, eps=layer_norm_eps, bias=bias, **factory_kwargs)
        self.dropout1 = Dropout(dropout)
        self.dropout2 = Dropout(dropout)

        # Legacy string support for activation function.
        if isinstance(activation, str):
            activation = _get_activation_fn(activation)

        # We can't test self.activation in forward() in TorchScript,
        # so stash some information about it instead.
        if activation is F.relu or isinstance(activation, torch.nn.ReLU):
            self.activation_relu_or_gelu = 1
        elif activation is F.gelu or isinstance(activation, torch.nn.GELU):
            self.activation_relu_or_gelu = 2
        else:
            self.activation_relu_or_gelu = 0
        self.activation = activation

    def __setstate__(self, state):
        super().__setstate__(state)
        if not hasattr(self, "activation"):
            self.activation = F.relu

    def forward(
        self,
        src: Tensor,
        src_mask: Optional[Tensor] = None,
        src_key_padding_mask: Optional[Tensor] = None,
        is_causal: bool = False,
        output_attentions: bool = False,
    ) -> Tensor:
        r"""Pass the input through the encoder layer.

        Args:
            src: the sequence to the encoder layer (required).
            src_mask: the mask for the src sequence (optional).
            src_key_padding_mask: the mask for the src keys per batch (optional).
            is_causal: If specified, applies a causal mask as ``src mask``.
                Default: ``False``.
                Warning:
                ``is_causal`` provides a hint that ``src_mask`` is the
                causal mask. Providing incorrect hints can result in
                incorrect execution, including forward and backward
                compatibility.

        Shape:
            see the docs in :class:`~torch.nn.Transformer`.
        """
        src_key_padding_mask = F._canonical_mask(
            mask=src_key_padding_mask,
            mask_name="src_key_padding_mask",
            other_type=F._none_or_dtype(src_mask),
            other_name="src_mask",
            target_type=src.dtype,
        )

        src_mask = F._canonical_mask(
            mask=src_mask,
            mask_name="src_mask",
            other_type=None,
            other_name="",
            target_type=src.dtype,
            check_other=False,
        )

        is_fastpath_enabled = torch.backends.mha.get_fastpath_enabled()

        why_not_sparsity_fast_path = ""
        if not is_fastpath_enabled:
            why_not_sparsity_fast_path = (
                "torch.backends.mha.get_fastpath_enabled() was not True"
            )
        elif not src.dim() == 3:
            why_not_sparsity_fast_path = (
                f"input not batched; expected src.dim() of 3 but got {src.dim()}"
            )
        elif self.training:
            why_not_sparsity_fast_path = "training is enabled"
        elif not self.self_attn.batch_first:
            why_not_sparsity_fast_path = "self_attn.batch_first was not True"
        elif self.self_attn.in_proj_bias is None:
            why_not_sparsity_fast_path = "self_attn was passed bias=False"
        elif not self.self_attn._qkv_same_embed_dim:
            why_not_sparsity_fast_path = "self_attn._qkv_same_embed_dim was not True"
        elif not self.activation_relu_or_gelu:
            why_not_sparsity_fast_path = "activation_relu_or_gelu was not True"
        elif not (self.norm1.eps == self.norm2.eps):
            why_not_sparsity_fast_path = "norm1.eps is not equal to norm2.eps"
        elif src.is_nested and (
            src_key_padding_mask is not None or src_mask is not None
        ):
            why_not_sparsity_fast_path = "neither src_key_padding_mask nor src_mask are not supported with NestedTensor input"
        elif self.self_attn.num_heads % 2 == 1:
            why_not_sparsity_fast_path = "num_head is odd"
        elif torch.is_autocast_enabled():
            why_not_sparsity_fast_path = "autocast is enabled"
        elif any(
            len(getattr(m, "_forward_hooks", {}))
            + len(getattr(m, "_forward_pre_hooks", {}))
            for m in self.modules()
        ):
            why_not_sparsity_fast_path = "forward pre-/hooks are attached to the module"
        if not why_not_sparsity_fast_path:
            tensor_args = (
                src,
                self.self_attn.in_proj_weight,
                self.self_attn.in_proj_bias,
                self.self_attn.out_proj.weight,
                self.self_attn.out_proj.bias,
                self.norm1.weight,
                self.norm1.bias,
                self.norm2.weight,
                self.norm2.bias,
                self.linear1.weight,
                self.linear1.bias,
                self.linear2.weight,
                self.linear2.bias,
            )

            # We have to use list comprehensions below because TorchScript does not support
            # generator expressions.
            _supported_device_type = [
                "cpu",
                "cuda",
                torch.utils.backend_registration._privateuse1_backend_name,
            ]
            if torch.overrides.has_torch_function(tensor_args):
                why_not_sparsity_fast_path = "some Tensor argument has_torch_function"
            elif not all(
                (x.device.type in _supported_device_type) for x in tensor_args
            ):
                why_not_sparsity_fast_path = (
                    "some Tensor argument's device is neither one of "
                    f"{_supported_device_type}"
                )
            elif torch.is_grad_enabled() and any(x.requires_grad for x in tensor_args):
                why_not_sparsity_fast_path = (
                    "grad is enabled and at least one of query or the "
                    "input/output projection weights or biases requires_grad"
                )

            if not why_not_sparsity_fast_path:
                merged_mask, mask_type = self.self_attn.merge_masks(
                    src_mask, src_key_padding_mask, src
                )
                return torch._transformer_encoder_layer_fwd(
                    src,
                    self.self_attn.embed_dim,
                    self.self_attn.num_heads,
                    self.self_attn.in_proj_weight,
                    self.self_attn.in_proj_bias,
                    self.self_attn.out_proj.weight,
                    self.self_attn.out_proj.bias,
                    self.activation_relu_or_gelu == 2,
                    self.norm_first,
                    self.norm1.eps,
                    self.norm1.weight,
                    self.norm1.bias,
                    self.norm2.weight,
                    self.norm2.bias,
                    self.linear1.weight,
                    self.linear1.bias,
                    self.linear2.weight,
                    self.linear2.bias,
                    merged_mask,
                    mask_type,
                )

        # see Fig. 1 of https://arxiv.org/pdf/2002.04745v1.pdf
        x = src
        if self.norm_first:
            if output_attentions:
                x, attn_map = x + self._sa_block(
                    self.norm1(x),
                    src_mask,
                    src_key_padding_mask,
                    is_causal=is_causal,
                    output_attentions=output_attentions,
                )
            else:
                x = x + self._sa_block(
                    self.norm1(x),
                    src_mask,
                    src_key_padding_mask,
                    is_causal=is_causal,
                )
            x = x + self._ff_block(self.norm2(x))
        else:
            if output_attentions:
                attn_output, attn_map = self._sa_block(
                    x,
                    src_mask,
                    src_key_padding_mask,
                    is_causal=is_causal,
                    output_attentions=output_attentions,
                )
                x = self.norm1(x + attn_output)
            else:
                x = self.norm1(
                    x
                    + self._sa_block(
                        x,
                        src_mask,
                        src_key_padding_mask,
                        is_causal=is_causal,
                    )
                )
            x = self.norm2(x + self._ff_block(x))
        if output_attentions:
            return x, attn_map
        else:
            return x

    # self-attention block
    def _sa_block(
        self,
        x: Tensor,
        attn_mask: Optional[Tensor],
        key_padding_mask: Optional[Tensor],
        is_causal: bool = False,
        output_attentions: bool = False,
    ) -> Tensor:
        if output_attentions:
            # If output_attentions is True, we return the attention map
            x, attn_map = self.self_attn(
                x,
                x,
                x,
                attn_mask=attn_mask,
                key_padding_mask=key_padding_mask,
                need_weights=True,
                is_causal=is_causal,
                average_attn_weights=False,
            )
            return self.dropout1(x), attn_map
        else:
            x = self.self_attn(
                x,
                x,
                x,
                attn_mask=attn_mask,
                key_padding_mask=key_padding_mask,
                need_weights=False,
                is_causal=is_causal,
            )[0]
            return self.dropout1(x)

    # feed forward block
    def _ff_block(self, x: Tensor) -> Tensor:
        x = self.linear2(self.dropout(self.activation(self.linear1(x))))
        return self.dropout2(x)


def _get_clones(module, N):
    # FIXME: copy.deepcopy() is not defined on nn.module
    return ModuleList([copy.deepcopy(module) for i in range(N)])


def _get_activation_fn(activation: str) -> Callable[[Tensor], Tensor]:
    if activation == "relu":
        return F.relu
    elif activation == "gelu":
        return F.gelu

    raise RuntimeError(f"activation should be relu/gelu, not {activation}")


def _detect_is_causal_mask(
    mask: Optional[Tensor],
    is_causal: Optional[bool] = None,
    size: Optional[int] = None,
) -> bool:
    """Return whether the given attention mask is causal.

    Warning:
    If ``is_causal`` is not ``None``, its value will be returned as is.  If a
    user supplies an incorrect ``is_causal`` hint,

    ``is_causal=False`` when the mask is in fact a causal attention.mask
       may lead to reduced performance relative to what would be achievable
       with ``is_causal=True``;
    ``is_causal=True`` when the mask is in fact not a causal attention.mask
       may lead to incorrect and unpredictable execution - in some scenarios,
       a causal mask may be applied based on the hint, in other execution
       scenarios the specified mask may be used.  The choice may not appear
       to be deterministic, in that a number of factors like alignment,
       hardware SKU, etc influence the decision whether to use a mask or
       rely on the hint.
    ``size`` if not None, check whether the mask is a causal mask of the provided size
       Otherwise, checks for any causal mask.
    """
    # Prevent type refinement
    make_causal = is_causal is True

    if is_causal is None and mask is not None:
        sz = size if size is not None else mask.size(-2)
        causal_comparison = _generate_square_subsequent_mask(
            sz, device=mask.device, dtype=mask.dtype
        )

        # Do not use `torch.equal` so we handle batched masks by
        # broadcasting the comparison.
        if mask.size() == causal_comparison.size():
            make_causal = bool((mask == causal_comparison).all())
        else:
            make_causal = False

    return make_causal



import torch
from .model_dir import TransformerModel
from helical.models.scgpt.scgpt_config import scGPTConfig
import json
from typing import List, Mapping, Optional
import torch
import logging

LOGGER = logging.getLogger(__name__)


def load_pretrained(
    model: torch.nn.Module,
    pretrained_params: Mapping[str, torch.Tensor],
    strict: bool = False,
    prefix: Optional[List[str]] = None,
    verbose: bool = True,
) -> torch.nn.Module:
    """
    Load pretrained weights to the model.

    Args:
        model (torch.nn.Module): The model to load weights to.
        pretrained_params (Mapping[str, torch.Tensor]): The pretrained parameters.
        strict (bool): Whether to strictly enforce that the keys in :attr:`pretrained_params`
            match the keys returned by this module's :meth:`Module.state_dict`. Default to False.
        prefix (List[str]): The list of prefix strings to match with the keys in
            :attr:`pretrained_params`. The matched keys will be loaded. Default to None.

    Returns:
        torch.nn.Module: The model with pretrained weights.
    """

    # as we do not use fast transformer, we need to rename the keys
    pretrained_params = {
        k.replace("Wqkv.", "in_proj_"): v for k, v in pretrained_params.items()
    }

    if prefix is not None and len(prefix) > 0:
        if isinstance(prefix, str):
            prefix = [prefix]
        pretrained_params = {
            k: v
            for k, v in pretrained_params.items()
            if any(k.startswith(p) for p in prefix)
        }

    model_dict = model.state_dict()
    if strict:
        if verbose:
            for k, v in pretrained_params.items():
                LOGGER.info(f"Loading parameter {k} with shape {v.shape}")
        model_dict.update(pretrained_params)
        model.load_state_dict(model_dict)
    else:
        if verbose:
            for k, v in pretrained_params.items():
                if k in model_dict and v.shape == model_dict[k].shape:
                    LOGGER.info(f"Loading parameter {k} with shape {v.shape}")
        pretrained_params = {
            k: v
            for k, v in pretrained_params.items()
            if k in model_dict and v.shape == model_dict[k].shape
        }
        model_dict.update(pretrained_params)
        model.load_state_dict(model_dict)

    return model


def load_model(model_configs: scGPTConfig):

    # load model and vocabulary
    model_dir = model_configs["model_path"].parent
    vocab_file = model_dir / "vocab.json"

    # vocabulary
    with vocab_file.open("r") as f:
        vocab = json.load(f)

    model = TransformerModel(
        ntoken=len(vocab),
        d_model=model_configs["embsize"],
        nhead=model_configs["nheads"],
        d_hid=model_configs["d_hid"],
        nlayers=model_configs["nlayers"],
        nlayers_cls=model_configs["n_layers_cls"],
        n_cls=1,
        vocab=vocab,
        dropout=model_configs["dropout"],
        pad_token=model_configs["pad_token"],
        pad_value=model_configs["pad_value"],
        do_mvc=True,
        do_dab=False,
        use_batch_labels=False,
        domain_spec_batchnorm=False,
        explicit_zero_prob=False,
        pre_norm=False,
    )

    load_pretrained(
        model,
        torch.load(model_configs["model_path"], map_location=model_configs["device"]),
        verbose=False,
    )
    model.to(model_configs["device"])
    model.eval()
    return model, vocab



def pad_mask_factory(mask):
    def pad_mask(b, h, q_idx, kv_idx):
        if mask is None:
            return True
        m = mask[b, kv_idx]
        return m

    return pad_mask


def causal_mask_factory(start_row=0, start_col=0, offset=0):
    def causal_mask(b, h, q_idx, kv_idx):
        m = (q_idx >= kv_idx + offset) & (q_idx >= start_row) & (kv_idx >= start_col)
        return m

    return causal_mask



import os
from typing import Any, Literal

import anndata as ad
from scanpy.readwrite import _check_datafile_present_and_download

PathLike = os.PathLike | str


def tabula_sapiens(
    tissue: Literal[
        "lymphnode",
        "heart",
        "ear",
        "endothelium",
        "epithelium",
        "germline",
        "immune",
        "neural",
        "stromal",
        "bladder",
        "blood",
        "bone_marrow",
        "ear",
        "eye",
        "fat",
        "heart",
        "kidney",
        "large_intestine",
        "liver",
        "lung",
        "lymph_node",
        "mammary",
        "muscle",
        "ovary",
        "pancreas",
        "prostate",
        "salivary_gland",
        "skin",
        "small_intestine",
        "spleen",
        "stomach",
        "testis",
        "thymus",
        "tongue",
        "trachea",
        "uterus",
        "vasculature",
    ],
    path: PathLike | None = None,
    force_download: bool = False,
    version: Literal["v1", "v2"] = "v2",
    **kwargs: Any,
) -> ad.AnnData:
    """
    Load tissue dataset from Tabula Sapiens.

    Args:
        tissue: Tissue name e.g. ('lymphnode', 'heart', or 'ear')
        path: Path to save the dataset. If None, uses default path.
        force_download: Whether to force download the dataset.
        version: Version of the dataset. 'v1' is in-distribution for Transcriptformer,
                'v2' is out-of-distribution.

    Returns
    -------
        AnnData object.
    """
    urls = {
        "endothelium": "https://datasets.cellxgene.cziscience.com/de698978-3267-45c6-b492-4b0636ef564d.h5ad",
        "epithelium": "https://datasets.cellxgene.cziscience.com/f4f99331-40cf-4bc7-bbb5-c4dc4c150840.h5ad",
        "germline": "https://datasets.cellxgene.cziscience.com/d81eae50-e96d-44a6-876a-540a3a4610dd.h5ad",
        "immune": "https://datasets.cellxgene.cziscience.com/7f2e355e-9944-4477-98f0-2184943d9d2c.h5ad",
        "neural": "https://datasets.cellxgene.cziscience.com/74d7bc5e-ac01-43d6-ae70-8ffe82b8afde.h5ad",
        "stromal": "https://datasets.cellxgene.cziscience.com/1c63b188-d097-441f-8922-b7ebaf715cca.h5ad",
        "bladder": "https://datasets.cellxgene.cziscience.com/01ef7ded-020c-4292-b253-a234f7331757.h5ad",
        "blood": "https://datasets.cellxgene.cziscience.com/4eb58518-23ce-43e8-89ca-9e9fa87c081d.h5ad",
        "bone_marrow": "https://datasets.cellxgene.cziscience.com/bbf12ff4-7f22-4cf5-83cf-9cd37ce4d6f4.h5ad",
        "ear": "https://datasets.cellxgene.cziscience.com/51233375-325e-4e8e-b05d-208fd8cc4ba4.h5ad",
        "eye": "https://datasets.cellxgene.cziscience.com/a77c03ba-0748-4382-94bd-e1c358175d5a.h5ad",
        "fat": "https://datasets.cellxgene.cziscience.com/4ccbfe8d-0e59-4433-b53c-55a717a32d16.h5ad",
        "heart": "https://datasets.cellxgene.cziscience.com/832cf70d-ffdd-48c1-b2cc-ffbb30b42e6c.h5ad",
        "kidney": "https://datasets.cellxgene.cziscience.com/b7581463-9676-46fd-9b84-73132109b2a4.h5ad",
        "large_intestine": "https://datasets.cellxgene.cziscience.com/8a18fafb-9d37-4bae-9a18-00e35368ca23.h5ad",
        "liver": "https://datasets.cellxgene.cziscience.com/c264e09f-7c3b-4294-b0f4-82a790bd0014.h5ad",
        "lung": "https://datasets.cellxgene.cziscience.com/7c5ee207-950a-4ae3-bd85-2483f75111d7.h5ad",
        "lymph_node": "https://datasets.cellxgene.cziscience.com/b439da44-6128-483f-9424-10c5618a8b7e.h5ad",
        "mammary": "https://datasets.cellxgene.cziscience.com/c207bca6-a5c8-472e-be14-83b7b8cc83a6.h5ad",
        "muscle": "https://datasets.cellxgene.cziscience.com/9bea986b-f7f1-479b-92d9-61a893629399.h5ad",
        "ovary": "https://datasets.cellxgene.cziscience.com/6e4ccce5-532e-401e-b1bf-ce01f4883eff.h5ad",
        "pancreas": "https://datasets.cellxgene.cziscience.com/240e2b84-6893-4651-830b-5baebe8b8bfe.h5ad",
        "prostate": "https://datasets.cellxgene.cziscience.com/250cb362-68b4-426e-b4b6-1f7c478197d2.h5ad",
        "salivary_gland": "https://datasets.cellxgene.cziscience.com/ded561fb-8668-4ed9-b239-f03a01584dda.h5ad",
        "skin": "https://datasets.cellxgene.cziscience.com/c458ba64-399b-4afb-b4b9-8ac91b8f43b3.h5ad",
        "small_intestine": "https://datasets.cellxgene.cziscience.com/a77da5e7-8a67-4351-9ec3-49a9c80060ab.h5ad",
        "spleen": "https://datasets.cellxgene.cziscience.com/79dd0749-aba8-40b8-80b8-2937377c2bdd.h5ad",
        "stomach": "https://datasets.cellxgene.cziscience.com/70b2a188-2183-4b20-80e2-8c870e6b01f0.h5ad",
        "testis": "https://datasets.cellxgene.cziscience.com/7ffe124c-b1ac-4998-932c-43e2038aacf9.h5ad",
        "thymus": "https://datasets.cellxgene.cziscience.com/3c5738bc-da1e-4b98-a2a6-685a7748e747.h5ad",
        "tongue": "https://datasets.cellxgene.cziscience.com/9e6d1c2e-b77d-4b36-b747-fad671e245f0.h5ad",
        "trachea": "https://datasets.cellxgene.cziscience.com/f950a2b4-0761-4ec3-b74e-4d951e2fe30b.h5ad",
        "uterus": "https://datasets.cellxgene.cziscience.com/e53087ec-74a9-4a37-b5fb-36911c3a68c7.h5ad",
        "vasculature": "https://datasets.cellxgene.cziscience.com/866ad7f6-bf26-4da4-b078-1b5eebee0fd6.h5ad",
    }

    if path is None:
        path = f"~/.cache/transcriptformer/{tissue}_{version}.h5ad"

    adata = _load_dataset_from_url(
        path,
        file_type="h5ad",
        backup_url=urls[tissue],
        force_download=force_download,
        **kwargs,
    )

    return filter_anndata_by_tissue_and_version(adata, version=version)


def bgee_testis_evolution(
    organism: Literal[
        "marmoset",
        "rhesus_macaque",
        "human",
        "chimpanzee",
        "platypus",
        "mouse",
        "opossum",
        "gorilla",
        "chicken",
    ],
    path: PathLike | None = None,
    force_download: bool = False,
    **kwargs: Any,
) -> ad.AnnData:
    """
    Load testis tissue from an organism dataset from Bgee Evolution.

    Args:
        organism: Organism name e.g. ('marmoset', 'rhesus_macaque').
        path: Path to save the dataset. If None, uses default path.
        force_download: Whether to force download the dataset.

    Returns
    -------
        AnnData object.
    """
    urls = {
        "marmoset": "datasets/v1/evo_distance/testis/Callithrix_jacchus_ERP132588_droplet_based_curated.h5ad",
        "chicken": "datasets/v1/evo_distance/testis/Gallus_gallus_ERP132576_droplet_based_curated.h5ad",
        "gorilla": "datasets/v1/evo_distance/testis/Gorilla_gorilla_ERP132581_droplet_based_curated.h5ad",
        "human": "datasets/v1/evo_distance/testis/Homo_sapiens_ERP132584_droplet_based_curated.h5ad",
        "macaque": "datasets/v1/evo_distance/testis/Macaca_mulatta_ERP132582_droplet_based_curated.h5ad",
        "opossum": "datasets/v1/evo_distance/testis/Monodelphis_domestica_ERP132579_droplet_based_curated.h5ad",
        "mouse": "datasets/v1/evo_distance/testis/Mus_musculus_ERP132577_droplet_based_curated.h5ad",
        "platypus": "datasets/v1/evo_distance/testis/Ornithorhynchus_anatinus_ERP132578_droplet_based_curated.h5ad",
        "chimpanzee": "datasets/v1/evo_distance/testis/Pan_troglodytes_ERP139683_droplet_based_curated.h5ad",
    }

    if path is None:
        path = f"~/.cache/transcriptformer/{organism}.h5ad"

    adata = _load_dataset_from_url(
        path,
        file_type="h5ad",
        backup_url=urls[organism],
        bucket_name="cz-benchmarks-data",
        force_download=force_download,
        **kwargs,
    )

    return adata


def download_all_embeddings(
    path: PathLike | None = None,
    force_download: bool = False,
    **kwargs: Any,
) -> None:
    """
    Download all embeddings used in the Transcriptformer paper.

    Args:
        path: Path to save the dataset. If None, uses default path.
        force_download: Whether to force download the dataset and overwrite existing files.
        **kwargs: Additional arguments passed to the download function.
    """
    import shutil
    import tarfile

    import boto3
    from botocore import UNSIGNED
    from botocore.config import Config

    file_name = "all_embeddings.tar.gz"

    if path is None:
        path = "~/.cache/transcriptformer"

    # Ensure path is expanded and normalized
    path = os.path.expanduser(path)

    # Define the tar file path and extraction directory
    tar_path = os.path.join(path, file_name)
    extraction_dir = path  # Extract to the same directory as the tar file

    bucket_name = "czi-transcriptformer"
    backup_url = "weights/" + file_name
    s3_client = boto3.client("s3", config=Config(signature_version=UNSIGNED))

    # Create directory if it doesn't exist
    os.makedirs(path, exist_ok=True)

    # If force_download is True, remove existing files
    if force_download:
        # Remove extracted files if they exist
        for item in os.listdir(extraction_dir):
            item_path = os.path.join(extraction_dir, item)
            if (
                os.path.isdir(item_path)
                and item != os.path.basename(file_name).split(".")[0]
            ):
                shutil.rmtree(item_path)
            elif os.path.isfile(item_path) and item != file_name:
                os.remove(item_path)

        # Remove tar file if it exists
        if os.path.exists(tar_path):
            os.remove(tar_path)

    # Download the file if it doesn't exist
    if not os.path.exists(tar_path):
        s3_client.download_file(bucket_name, backup_url, tar_path)

    # Extract the tar file
    with tarfile.open(tar_path, "r:gz") as tar:
        for member in tar.getmembers():
            member_path = os.path.abspath(os.path.join(extraction_dir, member.name))
            if not member_path.startswith(os.path.abspath(extraction_dir)):
                raise ValueError(f"Illegal tar archive entry: {member.name}")
            if member.islnk() or member.issym():
                raise ValueError(
                    f"Unsupported symbolic link in tar archive: {member.name}"
                )
        tar.extractall(
            extraction_dir,
            members=[
                member
                for member in tar.getmembers()
                if os.path.abspath(
                    os.path.join(extraction_dir, member.name)
                ).startswith(os.path.abspath(extraction_dir))
            ],
        )


def _load_dataset_from_url(
    fpath: PathLike,
    file_type: Literal["h5ad"],
    *,
    backup_url: str,
    bucket_name: str | None = None,
    force_download: bool = False,
    **kwargs: Any,
) -> ad.AnnData:
    fpath = os.path.expanduser(fpath)
    assert file_type in [
        "h5ad"
    ], f"Invalid type `{file_type}`. Must be one of `['h5ad']`."
    if not fpath.endswith(file_type):
        fpath += f".{file_type}"
    if force_download and os.path.exists(fpath):
        os.remove(fpath)
    if bucket_name is not None:
        import boto3
        from botocore import UNSIGNED
        from botocore.config import Config

        s3_client = boto3.client("s3", config=Config(signature_version=UNSIGNED))
        if not os.path.exists(fpath):
            os.makedirs(os.path.dirname(fpath), exist_ok=True)
            # Download the file
            s3_client.download_file(bucket_name, backup_url, fpath)
    else:
        if not _check_datafile_present_and_download(backup_url=backup_url, path=fpath):
            raise FileNotFoundError(f"File `{fpath}` not found or download failed.")
    data = ad.read_h5ad(filename=fpath, **kwargs)

    return data


def filter_anndata_by_tissue_and_version(
    adata: ad.AnnData,
    version: Literal["v1", "v2"],
    min_gene_counts: int = 10,
) -> ad.AnnData:
    if version == "v1":
        mask = adata.obs["donor_id"].str.split("TSP").str[-1].astype(int) < 16
    elif version == "v2":
        mask = adata.obs["donor_id"].str.split("TSP").str[-1].astype(int) > 16
    else:
        raise ValueError(f"Invalid version: {version}. Must be one of ['v1', 'v2']")

    adata_filtered = adata[mask].copy()

    gene_sums = adata_filtered.X.sum(axis=0)
    genes_to_keep = gene_sums >= min_gene_counts
    adata_filtered = adata_filtered[:, genes_to_keep].copy()

    return adata_filtered


if __name__ == "__main__":
    download_all_embeddings()



from .model import (
    TransformerModel,
    FlashTransformerEncoderLayer,
    GeneEncoder,
    AdversarialDiscriminator,
    MVCDecoder,
)
from .multiomic_model import MultiOmicTransformerModel
from .dsbn import *
from .grad_reverse import *
